#!/usr/bin/python
# -*- coding: utf-8 -*-

#############################################################################
# Version: 2 (Jul, 2014)                                                    #
# Author: Roberta Montefusco (rrt.montefusco@gmail.com), University of Pisa #
#                                                                           # 
# Generates propagated dependency file from CoNLL SD format file.           #
#############################################################################


"""
NAME:
  SDtoRels.py - generates triples dependency from CoNLL SD format file.
	
SYNOPSIS:
  ./SDtoRels.py [-p] [-o output_file] input_file

DESCRIPTION:
  This script gets an input file in CoNLL format, written using Stanford
  Dependencies (SD) notation and yields an output file, written using the
  same notation in "rels" format: dependency(governor-i, dependent-j),
  where i,j are the words positions inside the sentence.
	
  It returns also a file, STDerrors, with all errors and warnings found.
	
ELABORATION OPTIONS:
  -p	Return the "collapsed and propagated" representation.

OTHER OPTIONS:
  -d	Report debug informations in STDreports.
  -h	Output this manual and terminates.

NOTES:
  More informations on:
  - SD: http://nlp.stanford.edu/software/dependencies_manual.pdf.
  - CoNLL: http://acl.ldc.upenn.edu/W/W06/W06-2920.pdf.
  
#############################################################################
# Version: 2 (Jul, 2014)                                                    #
# Author: Roberta Montefusco (rrt.montefusco@gmail.com), University of Pisa #
#############################################################################
"""

import getopt
import sys
import codecs
import os

		

# variabile dello sviluppatore
DEBUG = False

reportFileName = "STDreports"

freport = ""



# nome del file di errore, file e messaggi di errore
errorFileName = "STDerrors"

ferr = ""

errors = ["WARNING - \"pobj\" not found.", "WARNING - relative sentence.", \
					"WARNING - found a \"prep\" under another \"prep\".", "WARNING - found \"prep\" under an \"mwe\".", \
					"WARNING - cannot convert in copula structure: two dependencies with \"acomp\" or \"attr\".", \
					"WARNING - cannot convert in copula structure: two dependencies with \"nsubj\".", \
					"WARNING - cannot convert in copula structure: dependencies with \"amod\" in the children.",
					"WARNING - found children under a \"cc\" dependency."]



# lista di token individuati come disgiunzioni; tutti gli altri sono considerati congiunzioni.
conj_o = ['o', 'oppure', 'od']



# stampa mess sul file ferr e stampa il manuale se h = True
def exit(mess, h):
	# global ferr
	# ferr.write(mess)
	sys.stderr.write("\n" + mess)
	if h:
		# ferr.write(__doc__)
		sys.stderr.write(__doc__)
	sys.stderr.write("\n")
	if not ferr.tell():
		os.remove(errorFileName)
	ferr.close()
	sys.exit()	
	
	
	
# token 'form' in posizione 'n' nella frase, viene arricchito con informazioni per diventare
# nodo del grafo dove: in 'parents' ci sono le coppie (dipendenza, padre) [collegamento dal
# padre al nodo]; in 'children' ci sono le coppie (dipendenza, figlio) [collegamento dal nodo
# al figlio]; in 'copy' è segnato tramite apostrofo  se è un nodo che rappresenta la copia di un
# token della frase (uno per ogni numero di copia che se ne è fatta); in 'tag' e 'finetag' sono
# segnati i part-of-speech tag a grana grossa e a grana fine, più specifico; 'form', 'lemma' e
# 'morph' indicano la forma, il lemma e la morfologia; i campi 'notes' e 'prenotes' servono per
# passare informazioni aggiuntive; 'nline' memorizza la riga del file in cui compare il token.
class Token:
	def __init__(self, nline, items):
		self.n = int(items[0])
		self.form = items[1]
		self.lemma = items[2]
		self.tag = items[3]
		self.finetag = items[4]
		self.morph = items[5]
		self.parents = [Dep(items[7].lower(), int(items[6]))]
		self.prenotes = items[8]
		self.notes = items[9]
		self.children = []
		self.copy = ""
		self.visited = False
		self.nline = nline
	
	
	# due token sono uguali se hanno lo stesso campo form, lo stesso indice nella frase e sono lo
	# stesso numero di copia fatta
	def __eq__(self, other):
		return self.n == other.n and self.form == other.form and self.copy == other.copy
		
		
	# un token si pone in relazione ad un altro, sulla base dell'ordine numerico della posizione nella frase
	def __cmp__(self, other):
		return -1 if self.n < other.n else 1 if self.n > other.n else 0
		
		
		
# dipendenza 'dep' fra il nodo sottinteso e quello etichettato con 'id'
class Dep:
	def __init__(self, dep, id):
		self.dep = dep
		self.id = int(id)
	
	
	# non possono esistere due archi da un nodo in uno stesso nodo con diverse etichette, per cui
	# controllare l'id del nodo destinazione è sufficiente a stabilire l'uguaglianza
	def __eq__(self, other):
		return self.id == other.id
		
		
	# una dipendenza si pone in relazione ad un'altra, sulla base dell'ordine numerico dell'id	
	def __cmp__(self, other):
		return -1 if self.id < other.id else 1 if self.id > other.id else 0
		
		
	# restituisce la relazione della dipendenza
	def getDep(self):
		return self.dep
		
	
	# restituisce l'id della dipendenza
	def getId(self):
		return self.id
		
		
		
# rappresentazione del grafo come lista di nodi - in questa fase iniziale, la posizione
# di un token all'interno della frase è uguale all'id del token come nodo nel grafo, in quanto
# assegnati progressivamente a partire dal nodo di root cui corrisponde 0
class Sentence:	
	def __init__(self, tokens, punct):
		# aggiunge il token ROOT in testa all'elenco
		self.tokens = [Token(-1, [0, 'ROOT', 'ROOT', '_', '_', '_', -1, '_', '_', '_'])] + tokens
		for x in tokens:
			par_x = x.parents[0]
			if par_x.dep == "punct" and not punct:
				# elimina le dipendenze di punteggiatura (rif. manuale), rendendoli nodi sconnessi del grafo
				x.parents = []
			else:
				# inserisce il nodo nell'elenco dei figli del padre
				self.tokens[par_x.id].children.append(Dep(par_x.dep, x.n))


	# restituisce la posizione del token nella frase del nodo id
	def getPos(self, id):
		return self.tokens[id].n


	# restituisce il form del nodo id
	def getForm(self, id):
		return self.tokens[id].form
		
		
	# restituisce il lemma del nodo id
	def getLemma(self, id):
		return self.tokens[id].lemma
		
	
	# restituisce il tag a grana grossa del nodo id
	def getTag(self, id):
		return self.tokens[id].tag
		
	
	# restituisce il tag a grana fine del nodo id
	def getFineTag(self, id):
		return self.tokens[id].finetag


	# restituisce la morfologia del nodo id
	def getMorph(self, id):
		return self.tokens[id].morph


	# restituisce le pre-note segnate per il nodo
	def getPreNotes(self, id):
		return self.tokens[id].prenotes


	# restituisce le note segnate per il nodo
	def getNotes(self, id):
		return self.tokens[id].notes

	
	# restituisce	una stringa composta da apostrofi, rappresentante il numero di copie del nodo id
	def getCopy(self, id):
		return self.tokens[id].copy


	# restituisce un valore booleano che indica se un nodo è un nodo copia o meno	
	def isCopy(self, id):
		return self.getCopy(id) != ""


	# restituisce la posizione nella lista di nodi, del nodo passato come parametro
	def getIndex(self, node):
		return self.tokens.index(node) if node in self.tokens else -1
		
	
	# pone a False tutti i campi visited dei nodi
	def resetVisite(self):
		for x in self.tokens:
			x.visited = False
			
		
	# rappresentazione come tripla della frase usando la FORM
	def stringForm(self):
		s = ""
		# provare a mettere i nodi copia sotto i loro nodi originali
		for x in self.tokens[1:]:
			x.parents.sort()
			for rel in x.parents:
				s += ("%s(%s-%d%s, %s-%d%s)\n" % (rel.dep, self.getForm(rel.id), self.getPos(rel.id), self.getCopy(rel.id), x.form, x.n, x.copy)).encode('utf-8')
		return s + "\n"
		
		
	# rappresentazione come tripla della frase usando il LEMMA
	def stringLemma(self):
		s = ""
		# provare a mettere i nodi copia sotto i loro nodi originali
		for x in self.tokens[1:]:
			x.parents.sort()
			for rel in x.parents:
				s += ("%s(%s-%d%s, %s-%d%s)\n" % (rel.dep, self.getLemma(rel.id), self.getPos(rel.id), self.getCopy(rel.id), x.lemma, x.n, x.copy)).encode('utf-8')
		return s + "\n"
		
			 
	# aggiunge la dipendenza 'dep' da 'pid' a 'cid' come: (dipendenza, figlio) nella
	# lista dei figli del padre; (dipendenza, padre) nella lista dei genitori del figlio. 
	# l'arco ha sempre lo stesso orientamento, nel grafo.
	# se è data la posizione in cui si vuole inserirlo, inserisce la dipendenza prima di essa.
	def addDep(self, pid, dep, cid, pos = -1):
		if pid >= len(self.tokens) or pid == cid or pid < 0 or cid < 0 or cid >= len(self.tokens):
			return False
			
		if pos < 0:
			self.tokens[pid].children.append(Dep(dep, cid))
		else:
			self.tokens[pid].children.insert(pos, Dep(dep, cid))
		self.tokens[cid].parents.append(Dep(dep, pid))
		return True
		
		
	# rimuove la dipendenza 'dep' da 'pid' a 'cid' rimuovendo: l'arco
	# dal padre al figlio, eliminandolo dalla lista dei figli del padre; l'arco dal
	# figlio al padre, eliminandolo dalla lista dei genitori del figlio.
	# se non è indicato il figlio da scollegare, rimuove tutte le dipendenze del
	# padre, in ingresso e in uscita.
	def remDep(self, pid, dep = '', cid = -1):
		if pid >= len(self.tokens) or pid < 0 or cid >= len(self.tokens):
			return False
			
		if cid < 0:
			while self.tokens[pid].parents:
				self.remDep(self.tokens[pid].parents[0].id, '', pid)
			while self.tokens[pid].children:
				self.remDep(pid, '', self.tokens[pid].children[0].id)
		else:
			if Dep(dep, cid) not in self.tokens[pid].children or Dep(dep, pid) not in self.tokens[cid].parents:
				return False
			self.tokens[pid].children.remove(Dep(dep, cid))
			self.tokens[cid].parents.remove(Dep(dep, pid))
		return True
	
	
	# modifica la dipendenza 'dep' in 'newdep', da 'pid' a 'cid', modificando: 
	# l'arco dal padre al figlio, modificando la dipendenza dalla lista dei figli del padre;
	# l'arco dal figlio al padre, modificando la dipendenza dalla lista dei genitori del figlio.
	def modDepLabel(self, pid, dep, cid, newdep):
		if pid >= len(self.tokens) or pid < 0 or cid < 0 or cid >= len(self.tokens) or \
			Dep(dep, cid) not in self.tokens[pid].children or Dep(dep, pid) not in self.tokens[cid].parents:
			return False
		self.tokens[pid].children[self.tokens[pid].children.index(Dep(dep, cid))] = Dep(newdep, cid)
		self.tokens[cid].parents[self.tokens[cid].parents.index(Dep(dep, pid))] = Dep(newdep, pid)
		return True
		
		
	# modifica la dipendenza 'dep' da 'pid' a 'cid', ridirezionandola in 'newcid'. Rimuove
	# l'arco dal padre al vecchio figlio, aggiungendo quello dal padre al nuovo figlio.
	def modDepDest(self, pid, dep, cid, newcid):
		if pid >= len(self.tokens) or pid < 0 or cid < 0 or cid >= len(self.tokens) or newcid >= len(self.tokens) or \
		Dep(dep, cid) not in self.tokens[pid].children or Dep(dep, pid) not in self.tokens[cid].parents:
			return False
		
		self.tokens[pid].children[self.tokens[pid].children.index(Dep(dep, cid))] = Dep(dep, newcid)
		self.tokens[cid].parents.remove(Dep(dep, pid))
		self.tokens[newcid].parents.append(Dep(dep, pid))
		return True


	# trova e rimuove tutte le dipendenze 'neg' figlie del nodo id: ritorna True se ne ha
	# trovata almeno una, False altrimenti.
	def childNegative(self, id):
		if 'neg' in [link.dep for link in self.tokens[id].children]:
			for dest in [link.id for link in self.tokens[id].children if link.dep == 'neg']:
				self.remDep(id, 'neg', dest)
			return True
		return False
	
	
	# aggiunge nella lista dei nodi del grafo, un nodo copia di quello passato come parametro, il
	# numero di copia sarà indicato da 'ncopy'
	def addCopyNode(self, node, ncopy):
		self.tokens.append(Token(node.nline, [node.n, node.form, node.lemma, \
			node.tag, node.finetag, node.morph, -1, '_', '_', '_']))
		pos = len(self.tokens) - 1
		self.tokens[pos].parents = []
		self.tokens[pos].copy = "'" * ncopy
		return pos
		
	
	# sposta i figli aventi dipendenze diverse da quelle indicate in depsNotInL,
	# del nodo indicato come fr, facendoli diventare figli del nodo to
	def moveChildren(self, fr, to, depsNotInL = []):
		i = 0
		while i < len(self.tokens[fr].children):
			dep = self.tokens[fr].children[i].getDep()
			if len(depsNotInL) == 0 or dep not in depsNotInL:
				id = self.tokens[fr].children[i].getId()
				self.remDep(fr, dep, id)
				self.addDep(to, dep, id)
			else:
				i += 1
	
	
	# scrive l'errore iderr relativamente al nodo idnode su ferr
	def writeError(self, idnode, iderr):
		ferr.write("Line %d - " % self.tokens[idnode].nline + errors[iderr] + "\n")	
		
	
	# scrive il report nel file
	def writeReport(self, message, idnode = -1):
		if idnode != -1:
			freport.write("Line %d - " % self.tokens[idnode].nline)
		freport.write(message)
		
			
	# Gestisce collapse di preposizioni e congiunzioni.
	# restituisce un valore di verità e un valore numerico che indica il codice di errore generato:
	# - (True, -1,-1) : se non si sono verificati errori
	# - (False, id, i) : con i >= 0 se si è verificato un errore in corrispondenza del nodo id;
	#                    il messaggio corrisponde alla struttura globale errors[i].
	def collapse(self, node, agent_info, tree = False):
		# insieme di variabili che serve a monitorare le congiunzioni e i congiunti, trovati come
		# figli, e altre informazioni del nodo:
		# - conjs è la lista di tutti i congiunti trovati, rimasti in sospeso perché non è stata
		#   ancora trovata la congiunzione che li lega;
		# - firstCc è True fin quando non si incontra una congiunzione diversa dalla punteggiatura;
		# - latestCC è l'ultima congiunzione trovata (non negazione); è '' se non è stata ancora trovata;
		# - ncopy segna il numero di copie fatte del nodo;
		# - i indica la posizione del collegamento col figlio che si sta considerando.
		conjs = []
		firstCc = True
		latestCc = ''
		ncopy = 0
		i = 0
		
		nodeID = self.getIndex(node)
		
		# print "nodo: ", nodeID, [pippo.id for pippo in self.tokens[nodeID].children]
		
		while i < len(node.children):
			depChild = node.children[i].dep
			idChild = node.children[i].id
			i += 1
			
			# print "\t", [x.id for x in node.children], "\n\tfiglio: ", idChild, depChild
			
			# GESTIONE DELLE PREPOSIZIONI:
			# percorre l'arco quando la dipendenza è esattamente 'prep' - vuol dire che non è stata ancora toccata;
			# oppure quando è una sovrastringa 'prep_X' (che ha quindi già subito manipolazioni) da una multiword che
			# seguiva (quindi deve essere stata trovata una multiword) e non è un nodo copia (perché essi sono frutto
			# di passate elaborazioni e non bisogna ripercorrere le loro dipendenze).
			if depChild == 'prep' or (depChild.find('prep') == 0 and not self.isCopy(nodeID)):
				# insieme di variabili che serve a monitorare i collegamenti figli del nodo preposizione:
				# - conjs_copy è la lista di tutti gli ID dei nodi copia che si devono congiungere con il 
				#   nodo in esame, rimasti in sospeso perché non è stata ancora trovata la congiunzione;
				# - conjs_cc_pobj è la lista di tutte le coppie (ID, congiunzione), rimaste in sospeso,
				#   relative alle pobj o pcomp, figlie di preposizioni uguali a quella in esame, che devono
				#   essere collegate tramite congiunzione alla pobj della preposizione che si sta
				#   considerando. Le informazioni presenti in questa lista sono del tipo:
				#   - (ID, ''): se non è stata ancora trovata la congiunzione;
				#   - (ID, 'cc'): se la congiunzione è stata trovata ma non è stata trovata la pobj;
				# - conjs_prep_pobj è la lista di tutte le coppie (ID, 'prep_*') relative ai nodi che
				# 	devono collegarsi tramite la relazione indicata ad una pobj non ancora trovata;
				# - firstCc_prep e latestCc_prep assumono lo stesso significato delle rispettive firstCc e
				#   latestCc definite sopra;
				# - findPobj_firstPrep segnala se è stata trovata una pobj/pcomp per la preposizione in esame;
				# - composed_prep è una stringa che memorizza la preposizione composta trovata ad esclusione
				#   dell'ultima multiword trovata.
				# - have_seen_mwe assume valore vero se come figlio di una preposizione c'è una dipendenza mwe;
				# - pdep contiene l'indice della pobj figlia della preposizione; fin quando non la si trova
				#   ha valore -1;
				# - pdep_pobj è True se l'oggetto della preposizione è una pobj, False se è una pcomp;
				# - j indica la posizione del collegamento col figlio che si sta considerando.
				conjs_copy = []
				conjs_cc_pobj = []
				conjs_prep_pobj = []
				firstCc_prep = True
				latestCc_prep = ''
				findPobj_firstPrep = False
				composed_prep = ''
				pdep = -1
				pdep_pobj = True
				j = 0
				
				prep = self.getLemma(idChild).lower()
				
				# print "\t-> figli: ", [x.id for x in self.tokens[idChild].children]

				while j < len(self.tokens[idChild].children):
					depChildPrep = self.tokens[idChild].children[j].dep
					idChildPrep = self.tokens[idChild].children[j].id
					j += 1
					
					# print "\t\t", [x.id for x in self.tokens[idChild].children],"\n\t\tfiglio: ", idChildPrep, depChildPrep
					
					# GESTIONE DELLE PREPOSIZIONI:
					# NODO FIGLIO DELLA PREPOSIZIONE IN ESAME:
					# POBJ o PCOMP: creo il collegamento fra il padre della preposizione e l'oggetto di essa, con la dipendenza 
					#               appropriata; conservo poi l'indice per eventuali ulteriori collegamenti e memorizzo se è
					#               stata trovata una pcomp o una pobj, segnalo che è stata trovata quella della preposizione in esame,
					#               settando le opportune variabili booleane.
					if depChildPrep == 'pobj' or depChildPrep == 'pcomp':			
						# inizializzo la variabile che indica la presenza dell'agente: è vera solo se ho bisogno di quell'informazione
						# e il token in questione è un agente: in questo modo funziona sia che io non debba sfruttare l'informazione
						# sia che la debba sfruttare e non è un agente.
						isagent = (agent_info and self.getNotes(idChild) == 'AGENT')
							
						if depChild != 'prep':
							# se sto percorrendo una preposizione composta 'prep_X', estraggo la preposizione
							composed_prep = depChild[depChild.find('_') + 1: len(depChild)]
							preposition = 'agent' if isagent else 'prepc_' + composed_prep + '_' + prep if depChildPrep == 'pcomp' else 'prep_' + composed_prep + '_' + prep
						else:
							preposition = 'agent' if isagent else 'prepc_' + prep if depChildPrep == 'pcomp' else 'prep_' + prep
							
						self.addDep(nodeID, preposition, idChildPrep, i + len(conjs_cc_pobj) - 1)
						# print 'aggiunta: ', nodeID, preposition, idChildPrep, 'figli di nodeID: ', [aaa.id for aaa in self.tokens[nodeID].children]
						pdep = idChildPrep
						findPobj_firstPrep = True
						if depChildPrep == 'pcomp':
							pdep_pobj = False
					
					# GESTIONE DELLE PREPOSIZIONI:
					# NODO FIGLIO DELLA PREPOSIZIONE IN ESAME:
					# PREP: warning - non ci possono essere due preposizioni di seguito
					elif depChild == 'prep' and depChildPrep == 'prep':
						# print 'boom ', nodeID, idChild, idChildPrep
						# return False, idChild, 2
						# print "sono qui"
						self.writeError(idChild, 2)
						
					# GESTIONE DELLE PREPOSIZIONI:
					# NODO FIGLIO DELLA PREPOSIZIONE IN ESAME:
					# MWE: ci si trova di fronte ad una preposizione articolata su più parole, segnate come mwe.
					#      memorizza la preposizione composta da tutte le mwe che trova in catena (tranne l'ultima),
					#      scollegando i nodi man mano che li incontra. Alla fine, possono verificarsi due casi:
					#      - trova in fondo una pobj o una pcomp: rimuove il collegamento fra l'ultima mwe/prep trovata
					#        e la pobj/pcomp per creare il corrispondente collegamento fra il nodo iniziale considerato
					#        e l'oggetto della preposizione multipla, segnalo che è stata trovata la pobj/pcomp della
					#        preposizione in esame;
					#      - non trova in fondo nessuna pobj o pcomp: rinomina la dipendenza fra la preposizione e
					#        e l'ultima mwe/prep trovata nella preposizione composta ottenuta.
					elif depChildPrep == 'mwe':
						# in questo modo non avanza l'indice e si riesamina l'arco appena trovato
						i -= 1
						have_seen_mwe = True
						prep += '_' + self.getLemma(idChildPrep)
						l = [link.dep for link in self.tokens[idChildPrep].children]
						exID = idChildPrep
						
						while 'mwe' in l:
							index = l.index('mwe')
							child = self.tokens[idChildPrep].children[index].id
							prep += '_' + self.getLemma(child).lower()
							# rimuovo una dipendenza: tutti i figli diventano figli del nodo figlio di prep:
							# in questo modo, in fondo, diverranno figli del nodo di partenza			
							self.remDep(idChildPrep, 'mwe', child)
							self.moveChildren(idChildPrep, idChild)
							idChildPrep = child	
							l = [link.dep for link in self.tokens[idChildPrep].children]
						
						prep = prep[0:prep.rfind('_')]
						
						if 'prep' in l:
							# return False, idChildPrep, 3
							self.writeError(idChildPrep, 3)
						
						if 'pobj' in l or 'pcomp' in l:	
							nameDep = 'pobj' if 'pobj' in l else 'pcomp'
							nameRel = 'prep_' if 'pobj' in l else 'prepc_'
							self.addDep(nodeID, nameRel + prep, idChildPrep, i + len(conjs_cc_pobj))
						else:
							self.modDepLabel(idChild, '', exID, 'prep_' + prep)
					
					# GESTIONE DELLE PREPOSIZIONI:
					# NODO FIGLIO DELLA PREPOSIZIONE IN ESAME:
					# CC: elimina la dipendenza fra la preposizione e la congiunzione; se la congiunzione è un
					#     simbolo di punteggiatura, la ignora; altrimenti: segnala che non è la prima congiunzione
					#     trovata; la memorizza e provvede a sistemare tutti gli eventuali nodi rimasti pendenti:
					#     - aggiunge la dipendenza 'conj_*' fra il nodo in esame e le copie con la congiunzione trovata;
					#     - se l'oggetto della preposizione è stato già trovato, aggiunge le dipendenze 'conj_*'
					#       fra essa e tutte le pobj rimaste pendenti; altrimenti a tutte le pobj pendenti, segnala la
					#       congiunzione in attesa di essere collegati quando si troverà l'oggetto della preposizione.
					elif depChildPrep == 'cc':
						# in questo modo non avanza l'indice e si esamina il prossimo figlio che prenderà
						# il posto della 'cc' eliminata
						j -= 1
						cc_child = [link for link in self.tokens[idChildPrep].children]
						cc_not_mwe = [link.dep for link in cc_child if link.dep != 'mwe']
						if cc_not_mwe:
							self.writeError(idChildPrep, 7)
						if cc_child and DEBUG: # se attiva l'opzione di DEBUG, stampa cosa c'è sotto la congiunzione nel file di report
							self.writeReport(("The cc node \"%s-%d%s\" has these children:\n" % \
							(self.getForm(idChildPrep), self.getPos(idChildPrep), self.getCopy(idChildPrep))).encode('utf-8'), idChildPrep)
							for ccc in cc_child:
								self.writeReport(("%s(%s-%d%s, %s-%d%s)\n" % (ccc.dep, self.getForm(idChildPrep), self.getPos(idChildPrep), \
								self.getCopy(idChildPrep), self.getForm(ccc.id), self.getPos(ccc.id), self.getCopy(ccc.id))).encode('utf-8'))		
							self.writeReport("\n")
														
						self.remDep(idChild, 'cc', idChildPrep)
						if self.getTag(idChildPrep) != 'F':
							firstCc_prep = False
							latestCc_prep = 'o' if self.getForm(idChildPrep).encode('utf-8').lower() in conj_o else 'e'
							
							for id in conjs_copy:
								self.addDep(nodeID, 'conj_' + latestCc_prep, id, i + len(conjs_cc_pobj))
							conjs_copy = []
							
							if pdep >= 0:
								for id, cc in conjs_cc_pobj:
									self.addDep(pdep, 'conj_' + latestCc_prep, id)
								conjs_cc_pobj = []
							else:
								for i in range(len(conjs_cc_pobj)):
									conjs_cc_pobj[i] = (conjs_cc_pobj[i][0], latestCc_prep)
					
					# GESTIONE DELLE PREPOSIZIONI:
					# NODO FIGLIO DELLA PREPOSIZIONE IN ESAME:
					# CONJ: memorizza se il congiunto è negato; se ci si trova di fronte ad una preposizione
					#       articolata su più parole, segnate come mwe o prep, memorizza la preposizione
					#       composta da tutte le mwe o prep che trova in catena, scollegando i nodi man mano
					#       che li incontra; se la preposizione congiunto è diversa della preposizione padre,
					#       allora si crea un nodo copia e, se si dispone delle informazioni sulla congiunzione,
					#       si aggiunge la dipendenza 'conj_*' fra il nodo in esame e il nodo copia, altrimenti
					#       lo si inserisce nella lista, conjs_copy.
          #       Alla fine di questa fase, possono verificarsi due casi:
					#       - trova in fondo una pobj o una pcomp: rimuove il collegamento fra l'ultima mwe/prep
					#         trovata e la pobj/pcomp; 
					#         - se la preposizione padre e la preposizione congiunto sono uguali:        
					#           - se non è stata trovata una congiunzione: memorizza nella lista, conjs_cc_pobj,
					#             l'indice dell'oggetto della preposizione congiunto;
					#           - altrimenti aggiunge tale relazione in quanto disponde di tutte le informazioni;
					#         - se la preposizione padre e la preposizione congiunto sono diverse: aggiunge la
					#           dipendenza 'prep_*'/'prepc_*' fra il nodo copia e la pobj/pcomp del congiunto;
					#       - non trova in fondo nessuna pobj o pcomp: inserisce nella lista, conjs_prep_pobj,
					#         l'id del nodo (copia o originale a seconda di quello che si sta considerando) e
					#         la preposizione con cui si dovrà collegare.
					#       Rimuove infine tutti i collegamenti del congiunto.
					elif depChildPrep == 'conj':
						# in questo modo non avanza l'indice e si esamina il prossimo figlio che prenderà
						# il posto della 'conj' eliminata
						j -= 1
						# print "conj: ", idChildPrep
						neg_prep = self.childNegative(idChildPrep)
						prep_conj = self.getLemma(idChildPrep).lower()
						l = [link.dep for link in self.tokens[idChildPrep].children]
						exID = idChildPrep
						
						while 'mwe' in l:
							index = l.index('mwe')
							child = self.tokens[idChildPrep].children[index].id
							prep_conj += '_' + self.getLemma(child).lower()
							# rimuovo una dipendenza: tutti i figli diventano figli del nodo figlio di prep:
							# in questo modo, in fondo, diverranno figli del nodo di partenza			
							self.remDep(idChildPrep, 'mwe', child)
							self.moveChildren(idChildPrep, idChild)
							idChildPrep = child	
							l = [link.dep for link in self.tokens[idChildPrep].children]			
						
						if 'prep' in l:
							# return False, idChildPrep, 3
							self.writeError(idChildPrep, 3)
							
						if prep != prep_conj:
							ncopy += 1
							copyID = self.addCopyNode(node, ncopy)
							if neg_prep:
								self.addDep(nodeID, 'conj_negcc', copyID, i + len(conjs_cc_pobj))
							else:
								if firstCc_prep:
									conjs_copy += [copyID]
								else:
									self.addDep(nodeID, 'conj_' + latestCc_prep, copyID, i + ncopy + len(conjs_cc_pobj))
						
						nameDep = 'pobj' if 'pobj' in l else 'pcomp'
						nameRel = 'prep_' if 'pobj' in l else 'prepc_'
						
						if 'pobj' in l or 'pcomp' in l:
							index = l.index(nameDep)
							index = self.tokens[idChildPrep].children[index].id
							self.remDep(idChildPrep, nameDep, index)
							
							# print "rimuovo la dipendenza ", idChildPrep, nameDep, index
							if not findPobj_firstPrep:
								pdep = index
							if prep == prep_conj:
								if neg_prep:
									self.addDep(pdep, 'conj_negcc', index)
								else:
									if firstCc_prep:
										conjs_cc_pobj += [(index, '')]
									else:
										self.addDep(pdep, 'conj_' + latestCc_prep, index)
							else:
								if composed_prep != '':
									prep_conj = composed_prep + '_' + prep_conj
								self.addDep(copyID, nameRel + prep_conj, index, i + len(conjs_cc_pobj))
								# print "aggiunta la dipendenza ", copyID, nameRel + prep_conj, index
								# print "figli: ", [pluto.id for pluto in self.tokens[idChild].children]
						else:
							conjs_prep_pobj += [(copyID if ncopy else nodeID, prep_conj)]	
						
						self.moveChildren(exID, idChild, ['pobj', 'pcomp'])
						self.remDep(exID)
						
				# GESTIONE DELLE PREPOSIZIONI:
				# RISOLUZIONE DEI CONGIUNTI IN SOSPESO:		
				# si tratta di un elenco di congiunti separati da virgole senza una congiunzione finale,
				# che sarà quindi 'e'.
				for id, c in conjs_cc_pobj:
					if not self.addDep(pdep, 'conj_e', id):
						# return False, idChild, 0
						self.writeError(idChild, 0)
					
				for id in conjs_copy:
					self.addDep(nodeID, 'conj_e', id, i + len(conjs_cc_pobj))
				
				# GESTIONE DELLE PREPOSIZIONI:
				# RISOLUZIONE DEI CONGIUNTI IN SOSPESO:		
				# aggiunge tutte le dipendenze 'prep_*' in sospeso, una volta trovato l'oggetto della preposizione.	
				# se anche la preposizione in esame non ha come figli una pobj o una pcomp, controllando l'opportuna
				# variabile booleana, e sono state trovate congiunzioni (ci sono congiunti in sospeso o è stata trovata
				# la congiunzione), allora viene aggiunta alla lista delle dipendenze da completare.
				# se è stata selezionata la versione '-t' (tree = True), occorre preservare la struttura dell'albero,
				# per cui si crea un nodo copia dell'oggetto per ogni preposizione in sospeso.
				if conjs_cc_pobj or latestCc_prep:
					if not findPobj_firstPrep:
						conjs_prep_pobj += [(nodeID, prep)]
				
				init_prep = 'prep_' if pdep_pobj else 'prepc_'
				
				# se ci sono preposizioni in attesa di essere risolte ma non c'è un oggetto cui collegarsi, errore.
				if conjs_prep_pobj and pdep < 0:
					# return False, idChild, 0
					self.writeError(idChild, 0)
				
				safe = True
				ncopy = 0
				for id, dep in conjs_prep_pobj:
					if not tree:
						safe = self.addDep(id, init_prep + dep, pdep)
					else:
						ncopy += 1
						safe = self.addDep(id, init_prep + dep, self.addCopyNode(self.tokens[pdep], ncopy))
					if not safe:
						# return False, idChild, 0
						self.writeError(idChild, 0)

				# scollego la preposizione da tutti i nodi cui è collegato
				self.moveChildren(idChild, nodeID, ['pobj', 'pcomp'])
				self.remDep(idChild)
				# print 'alla fine: ', [ppp.id for ppp in self.tokens[nodeID].children]
			
			# GESTIONE DELL'AGENT:
			# AGENT: non effettua alcuna computazione ma si limita a controllare se fra i suoi figli c'è una relazione di
			#        preposizione e in tal caso lancia errore, in quanto un agent è pur sempre una preposizione e non può
			#        esserci una preposizione sotto un'altra preposizione.
			elif depChild == 'agent' and len([child.id for child in self.tokens[idChild].children if child.dep == 'prep']):
				# return False, idChild, 2
				self.writeError(idChild, 2)
				
			# GESTIONE DELLE CONGIUNZIONI:
			# CC: elimina la dipendenza fra il nodo e la congiunzione; se la congiunzione è un
			#     simbolo di punteggiatura, la ignora; altrimenti: segnala che non è la prima
			#     congiunzione trovata; la memorizza e provvede a sistemare tutti gli eventuali
			#     nodi rimasti pendenti.
			elif depChild == 'cc':
				# in questo modo non avanza l'indice e si esamina il prossimo figlio che prenderà
				# il posto della 'cc' eliminata
				i -= 1
				cc_child = [link for link in self.tokens[idChild].children]
				cc_not_mwe = [link.dep for link in cc_child if link.dep != 'mwe']
				if cc_not_mwe:
					self.writeError(idChild, 7)
				if cc_child and DEBUG: # se attiva l'opzione di DEBUG, stampa cosa c'è sotto la congiunzione nel file di report
					self.writeReport(("The cc node \"%s-%d%s\" has these children:\n" % \
					(self.getForm(idChild), self.getPos(idChild), self.getCopy(idChild))).encode('utf-8'), idChild)
					for ccc in cc_child:
						self.writeReport(("%s(%s-%d%s, %s-%d%s)\n" % (ccc.dep, self.getForm(idChild), self.getPos(idChild), \
						self.getCopy(idChild), self.getForm(ccc.id), self.getPos(ccc.id), self.getCopy(ccc.id))).encode('utf-8'))		
					self.writeReport("\n")
						
				self.remDep(nodeID, 'cc', idChild)
				if self.getTag(idChild) != 'F':
					firstCc = False
					latestCc = 'o' if self.getForm(idChild).encode('utf-8').lower() in conj_o else 'e'
					for c in conjs:
						self.modDepLabel(nodeID, 'conj', c, 'conj_' + latestCc)
					conjs = []
						
			# GESTIONE DELLE CONGIUNZIONI:
			# CONJ: quando trova un congiunto, se esso ha una dipendenza negativa come figlia
			#       allora si conclude modificando l'etichetta della congiunzione come negativa;
			#       altrimenti se non si conosce ancora la congiunzione, si inserisce l'ID del
			#       nodo in una lista in attesa di risolverlo; altrimenti si modifica l'etichetta
			#       della congiunzione con quella appropriata.
			elif depChild == 'conj':
				if self.childNegative(idChild):
					self.modDepLabel(nodeID, 'conj', idChild, 'conj_negcc')
				else:
					if firstCc:
						conjs += [idChild]
					else:
						self.modDepLabel(nodeID, 'conj', idChild, 'conj_' + latestCc)
			
		# TERMINE DELLA SCANSIONE DEI FIGLI DEL NODO:
		# RISOLUZIONE DEI CONGIUNTI IN SOSPESO:	
		# si tratta di un elenco di congiunti separati da virgole senza una congiunzione finale,
		# che sarà quindi 'e'.
		for c in conjs:
			self.modDepLabel(nodeID, 'conj', c, 'conj_e')

		return True, -1, -1
	
	
	# gestisce l'esplicitazione del soggetto esterno.
	# XSUBJ: aggiunge la dipendenza 'xsubj' dai figli 'xcomp' verso i figli 'nsubj' del nodo,
	# se il nodo 'xcomp' ha nelle note 'XSUBJ' oppure se esiste almeno un figlio del nodo
	# marcato come 'xcomp' che abbia nelle note 'XSUBJ'.
	def expl_xsubj(self, node):
		nsubjs = [n.id for n in node.children if n.dep == 'nsubj']
		xsubjs = []
		for x in node.children:
			if x.dep == 'xcomp' and (self.getNotes(x.id) == 'XSUBJ' or \
			'XSUBJ' in [self.getNotes(xchild.id) for xchild in self.tokens[x.id].children]):
				xsubjs += [x.id]
		for x in xsubjs:
			for n in nsubjs:
				self.addDep(x, 'xsubj', n)
	
	
	# gestisce normalizzazione di frasi relative.
	# restituisce un valore di verità e un valore numerico che indica il codice di errore generato:
	# - (True, -1,-1) : se non si sono verificati errori
	# - (False, id, i) : con i >= 0 se si è verificato un errore in corrispondenza del nodo id;
	#                    il messaggio corrisponde alla struttura globale errors[i].
	# GESTIONE DELLE RELATIVE:
	# RCMOD: cerca in tutto il sottoalbero (escludendo tutti quelli che hanno come
	#        dipendenza rcmod, perché sono altre relative), tutti i pronomi relativi 
	#        guardando che:
	#        - il campo finetag del nodo sia uguale a DR o PR e li inserisce in una lista;
	#        - il campo form del nodo sia 'dove' e la dipendenza sia 'advmod' e li
	#          inserisce in un'altra lista;
	#        Al termine della visita, per ogni pronome trovato, redireziona la sua dipendenza
	#        in ingresso verso la testa del nodo rcmod (le dipendenze di 'dove' vengono
	#        modificate in 'npadvmod') e scollega tutte le dipendenze con i figli.
	def coll_relative(self, node):
		nodeID = self.getIndex(node)
		
		# print nodeID, " figli: ", [(x.id, x.dep) for x in node.children]
		for child in node.children:
			if child.dep == 'rcmod':
				# print "relative di: ", nodeID, " con figlio: ", child.id
				rels = []
				adv = []
				q = [child.id] if child.id != nodeID else []
				while q:
					# print "lista q: ", q
					nID = q.pop(0)
					for c in self.tokens[nID].children:
						if self.getFineTag(c.id).upper() == 'DR' or self.getFineTag(c.id).upper() == 'PR':
							rels += [(nID, c.dep, c.id)]
						if self.getForm(c.id).lower() == 'dove' and c.dep == 'advmod':
							adv += [(nID, c.dep, c.id)]
					q += [c.id for c in self.tokens[nID].children if c.dep != 'rcmod']
					
				for f, dep, c in rels:
					if not self.modDepDest(f, dep, c, nodeID):
						# return False, child.id, 1
						self.writeError(child.id, 1)
					self.remDep(c)
					
				for f, dep, c in adv:
					if not self.modDepDest(f, dep, c, nodeID):
						# return False, child.id, 1
						self.writeError(child.id, 1)
					self.modDepLabel(f, dep, nodeID, 'npadvmod')
					self.remDep(c)
		
		return True, -1, -1		
		
	
	# gestisce esplicitazione di collegamento ref fra il pronome relativo e la testa di frasi relative.
	# GESTIONE DELLE RELATIVE:
	# RCMOD: cerca in tutto il sottoalbero (escludendo tutti quelli che hanno come
	#        dipendenza rcmod, perché sono altre relative), tutti i pronomi relativi 
	#        guardando che:
	#        - il campo finetag del nodo sia uguale a DR o PR e li inserisce in una lista;
	#        - il campo form del nodo sia 'dove' e la dipendenza sia 'advmod' e li
	#          inserisce in un'altra lista;
	#        Al termine della visita, per ogni pronome trovato, aggiunge una dipendenza
	#        di tipo 'ref' da esso verso la testa del nodo rcmod.
	def expl_relative(self, node):
		nodeID = self.getIndex(node)
		
		for child in node.children:
			if child.dep == 'rcmod':
				rels = []
				q = [child.id]
				while q:
					nID = q.pop(0)
					for c in self.tokens[nID].children:
						if self.getFineTag(c.id).upper() == 'DR' or self.getFineTag(c.id).upper() == 'PR' or \
							(self.getForm(c.id).lower() == 'dove' and c.dep == 'advmod'):
							rels += [c.id]
					q += [c.id for c in self.tokens[nID].children if c.dep != 'rcmod']	
				
				for r in rels:
					self.addDep(r, 'ref', nodeID)
			
	
	# gestisce la propagazione delle dipendenze sui congiunti.	
	def propagate(self, node):
		# trova tutte le etichette delle dipendenze con i suoi governor
		depsParents = [link.dep for link in node.parents]
		
		# trova tutte le etichette 'nsubj', 'conj*' delle dipendenze figlie e ne
		# memorizza l'id dei nodi figli
		conjuncts = []
		nsubjs = []
		for n in node.children:
			if n.dep.find('conj') == 0:
				conjuncts += [n.id]
			elif n.dep == 'nsubj' or n.dep == 'nsubjpass':
				nsubjs += [(n.id, n.dep)]
		
		# propaga le dipendenze dal padre del nodo ai figli del nodo se non ha una
		# dipendenza 'root' con il padre
		if not 'ROOT' in depsParents:
			for p in node.parents:
					for c in conjuncts:
						self.addDep(p.id, p.dep, c)
						
		# propaga le dipendenze dai figli del nodo con cui esso ha una relazione
		# 'conj' verso gli altri figli con cui ha una relazione 'nsubj' o 'nsubjpass'
		# se non esiste una relazione 'cop' fra il nodo e i suoi figli		
		if not 'cop' in depsParents:
			for c in conjuncts:
				for id, dep in nsubjs:
					self.addDep(c, dep, id)
				
	
	# esegue il collapse() su tutti i nodi
	def fun_coll(self, agent_info, tree = False):
		queque = [self.tokens[0].children[0].id]
		while queque:
			node = self.tokens[queque.pop(0)]
			node.visited = True
			result, errID, indexerr = self.collapse(node, agent_info, tree)
			if not result:
				self.writeError(errID, indexerr)
				return None
			queque += [x.id for x in node.children if not self.tokens[x.id].visited and x.id not in queque]
		self.resetVisite()
		
		
	# esegue il propagate() su tutti i nodi
	def fun_prop(self):
		queque = [self.tokens[0].children[0].id]
		while queque:
			node = self.tokens[queque.pop(0)]
			node.visited = True		
			self.propagate(node)
			queque += [x.id for x in node.children if not self.tokens[x.id].visited and x.id not in queque]
		self.resetVisite()
		
		
	# esegue il coll_relative() su tutti i nodi
	def fun_coll_relative(self):
		queque = [self.tokens[0].children[0].id]
		while queque:
			node = self.tokens[queque.pop(0)]
			node.visited = True	
			result, errID, indexerr = self.coll_relative(node)
			if not result:
				self.writeError(errID, indexerr)
				return None
			queque += [x.id for x in node.children if not self.tokens[x.id].visited and x.id not in queque]
		self.resetVisite()
		
		
	# esegue il expl_xsubj() su tutti i nodi
	def fun_expl_xsubj(self):
		queque = [self.tokens[0].children[0].id]
		while queque:
			node = self.tokens[queque.pop(0)]
			node.visited = True		
			self.expl_xsubj(node)
			queque += [x.id for x in node.children if not self.tokens[x.id].visited and x.id not in queque]
		self.resetVisite()
		
	
	# esegue il expl_relative su tutti i nodi
	def fun_expl_relative(self):
		queque = [self.tokens[0].children[0].id]
		while queque:
			node = self.tokens[queque.pop(0)]
			node.visited = True	
			result, errID, indexerr = self.expl_relative(node)
			if not result:
				self.writeError(errID, indexerr)
				return None
			queque += [x.id for x in node.children if not self.tokens[x.id].visited and x.id not in queque]
		self.resetVisite()
	
	
	# elabora le informazioni nella struttura: restituisce None se si sono verificati errori
	def convert(self, prop, coll, nocoll, tree, agent_info, xsubj_info, cop, ferr):
	# visita in ampiezza:
		if self.tokens[0].children:		
			
			# DEPRECATO - nel caso si dovesse riaggiungere, aggiungere l'opzione -e
			#if cop: # manipola la struttura dell'albero, prima di eseguire le successive operazioni
				#queque = [self.tokens[0].children[0].id]
				#while queque:
					#node = self.tokens[queque.pop(0)]
					#node.visited = True
					#node_id = self.getIndex(node)
					## se il lemma del nodo è "essere" e non ci sono dipendenze entranti di tipo "aux" o "auxpass"
					## ed esiste un figlio con dipendenza "nsubj" e un altro con dipendenza "attr" o "acomp"
					#deps_children = [x.dep for x in node.children]
					#deps_parents = [x.dep for x in node.parents]
					#if node.lemma.lower() == "essere" and "aux" not in deps_parents and "auxpass" not in deps_parents and \
					#"nsubj" in deps_children and ("attr" in deps_children or "acomp" in deps_children):
						#agg = [x for x in node.children if x.dep == "attr" or x.dep == "acomp"]
						#nsubj = [x for x in node.children if x.dep == "nsubj"]
						##if "amod" in deps_children:
							##self.writeError(node_id, 6)
							##break
						#if len(agg) != 1:
							#self.writeError(node_id, 4)
							#break
						#else:
							#agg = agg[0]
						#if len(nsubj) != 1:
							#self.writeError(node_id, 5)
							#break
						#else:
							#nsubj = nsubj[0]
						## sposta ogni dipendenza entrante in node.id, verso agg
						#for dep in node.parents:
							#self.modDepDest(dep.id, dep.dep, node_id, agg.id)
						## rimuovi la vecchia dipendenza "amod" o "attr" per poterla sostituire con una dipendenza
						## "cop" nell'altro verso
						#self.remDep(node_id, agg.dep, agg.id)
						#self.addDep(agg.id, "cop", node_id)
						## rimuovi la vecchia dipendenza "nsubj" per poterla sostituire con una dipendenza uguale ma
						## fra agg.id e il nodo in questione
						#self.remDep(node_id, "nsubj", nsubj.id)
						#self.addDep(agg.id, "nsubj", nsubj.id)
						#self.resetVisite()
						#queque = [self.tokens[0].children[0].id]
					#else:
						#queque += [x.id for x in node.children if not self.tokens[x.id].visited and x.id not in queque]
				#self.resetVisite()
				
			if prop: # per ogni nodo invoca: collapse() + coll_relative() + expl_xsubj() + propagate()
				self.fun_coll(agent_info)
				self.fun_coll_relative()
				if xsubj_info:
					self.fun_expl_xsubj()
				self.fun_prop()
					
			elif coll: # per ogni nodo invoca collapse() + coll_relative() + self.expl_xsubj()
				self.fun_coll(agent_info)
				self.fun_coll_relative()
				if xsubj_info:
					self.fun_expl_xsubj()
					
			elif nocoll: # per ogni nodo invoca expl_xsubj() + expl_relative()
				if xsubj_info:
					self.fun_expl_xsubj()
				self.fun_expl_relative()
				
			elif tree: # per ogni nodo invoca collapse()
				self.fun_coll(agent_info, tree)

		return self



def main():	
	prop = coll = nocoll = tree = cop = False
	lemma = agent_info = xsubj_info = punct = False
	tok = None
	outfile = ''
	global DEBUG
	global ferr
	ferr = codecs.open(errorFileName, 'w')
	
	try:
		opts, args = getopt.getopt(sys.argv[1:],'dho:p')
	except getopt.GetoptError:
		exit("Error in input options or arguments.\n\nFor info, see manual:\n", True)

	numOpts = 0	
	for opt, value in opts: # opts è una lista di coppie (opzione, valore)
	# ELAB
		if opt == '-p':
			prop = True
			numOpts += 1
	# OUT
		elif opt == '-o':
			outfile = value
	# OTHER
		elif opt == '-h':
			print __doc__
			if not ferr.tell():
				os.remove(errorFileName)
				ferr.close()
			sys.exit()	
		elif opt == '-d':
			DEBUG = True
		else:
			exit("Error in input options or arguments.\n\nFor info, see manual:\n", True)
	
	if numOpts > 1:
		exit("Error: too many options.\n\nFor info, see manual:\n", True)
				
	if len(args) == 1:
		infile = args[0]
	elif len(args) > 1:
		exit("Error: too many arguments.\n\nFor info, see manual:\n", True)
	else:
		exit("Error: not enough arguments.\n\nFor info, see manual:\n", True)
	
	try:
		fin = codecs.open(infile, 'r', 'utf-8')
	except:
		exit("Error while opening input file.\n", False)

	fout = codecs.open(outfile, 'w') if outfile else None
	if DEBUG:
		global freport
		freport = codecs.open(reportFileName, 'w')
	
	tokens = []
	nline = 1
	for line in fin:
		l = line.strip()
		
		if l:
			# crea un nuovo token con le informazioni rilevanti lette da file e lo inserisce in elenco
			items = l.split('\t')
			# controllo del formato del file: se errato termina il programma segnalando l'errore
			if len(items) != 10 or (not items[0].isdigit()) or (not items[6].isdigit()):
				exit("Error in the format file CoNLL at line %d.\n\nFor info, see manual:" % nline, True)
			tokens.append(Token(nline, items))
		else:
			# abbiamo una riga bianca quindi convertiamo la frase secondo i parametri
			result = Sentence(tokens, punct).convert(prop, coll, nocoll, tree, agent_info, xsubj_info, cop, ferr)
			result = "" if result == None else result.stringLemma() if lemma else result.stringForm()
			fout.write(result) if fout else sys.stdout.write(result)
			tokens = []
		
		nline += 1
		
	# nel caso in cui il file non termina con la riga vuota		
	if len(tokens) > 0: 
		result = Sentence(tokens, punct).convert(prop, coll, nocoll, tree, agent_info, xsubj_info, cop, ferr)
		result = "" if result == None else result.stringLemma() if lemma else result.stringForm()
		fout.write(result) if fout else sys.stdout.write(result)
	
	if not ferr.tell():
		ferr.write("No error was found.\n")
	ferr.close()
	fin.close()
	if outfile:
		fout.close()
				


if __name__ == '__main__':
    main()
